import typing as T
from enum import Enum
import random
import string
import re
from pydantic import BaseModel, Field


class FilterConnector(str, Enum):
    AND = "AND"
    OR = "OR"


class QueryBuilderError(Exception):
    pass


class ChildEdge(BaseModel):
    # for example, .artists
    db_expression: str | None
    qb: "QueryBuilder"


class QueryBuilder(BaseModel):
    typename: str | None = None
    fields: set[str] = Field(default_factory=set)
    variables: dict[str, T.Any] = Field(default_factory=dict)
    children: dict[str, ChildEdge] = Field(default_factory=dict)
    filter: str | None = None
    order_by: str | None = None
    offset: str | None = None
    limit: str | None = None

    full_query_str: str | None = None
    pattern_to_replace: str | None = None

    @staticmethod
    def build_child_var_name(
        child_name: str, var_name: str, variables_to_use: dict[str, T.Any]
    ) -> str:
        child_name = re.sub(r"[^a-zA-Z0-9]+", "_", child_name)
        count = 0
        while var_name in variables_to_use:
            count_str = "" if not count else f"_{count}"
            var_name = f"_{child_name}{count_str}_{var_name}"
        return var_name

    def build(self) -> tuple[str, dict[str, T.Any]]:
        variables_to_use = self.variables.copy()
        child_strs: set[str] = set()
        for child_name, child_edge in self.children.items():
            child = child_edge.qb
            child_str, child_variables = child.build()
            for var_name, var_val in child_variables.items():
                if var_name in variables_to_use:
                    if var_val is variables_to_use[var_name]:
                        continue
                    # must change the name for the child
                    new_var_name = self.build_child_var_name(
                        child_name=child_name,
                        var_name=var_name,
                        variables_to_use=variables_to_use,
                    )
                    variables_to_use[new_var_name] = var_val
                    # now, must regex the str to find this and replace it
                    regex = re.compile(r"\${}(?!\w)".format(var_name))
                    child_str = regex.sub(f"${new_var_name}", child_str)
                else:
                    variables_to_use[var_name] = var_val
            if not child_edge.db_expression:
                child_str = f"{child_name}: {child_str}"
            else:
                child_str = (
                    f"{child_name} := (select {child_edge.db_expression} {child_str})"
                )
            child_strs.add(child_str)

        fields_str = ", ".join([*sorted(self.fields), *sorted(child_strs)])
        s_parts = ["" if not fields_str else f"{{ {fields_str} }}"]
        if self.filter:
            s_parts.append(f"FILTER {self.filter}")
        if self.order_by:
            s_parts.append(f"ORDER BY {self.order_by}")
        if self.offset is not None:
            s_parts.append(self.offset)
        if self.limit is not None:
            s_parts.append(self.limit)
        final_s = " ".join(s_parts)
        if self.full_query_str:
            final_s = self.full_query_str.replace(self.pattern_to_replace, final_s)
        return final_s, variables_to_use

    def add_variable(
        self, key: str, val: T.Any, replace: bool = False
    ) -> "QueryBuilder":
        if key in self.variables:
            if not replace and self.variables[key] != val:
                raise QueryBuilderError(
                    f"Key {key} already exists in variables so you cannot add it. "
                    f"If you'd like to replace it, pass replace."
                )
        self.variables[key] = val
        return self

    def add_variables(
        self, variables: dict[str, T.Any], replace: bool = False
    ) -> "QueryBuilder":
        """if there is an error, it does not save to the builder"""
        if not variables:
            return self
        if not replace:
            for k in variables.keys():
                if k in self.variables:
                    raise QueryBuilderError(
                        f"Key {k} already exists in variables so you cannot add it. "
                        f"If you'd like to replace it, pass replace."
                    )
        self.variables.update(variables)
        return self

    # ADD HELPER FUNCTIONS FOR LIMIT AND OFFSET -> like add_offset...
    def set_offset(self, offset: int | None, replace: bool = False) -> "QueryBuilder":
        if offset is None:
            return self
        if self.offset is not None:
            if not replace:
                raise QueryBuilderError(
                    "An offset already exists. If you would like to replace it, pass in replace."
                )
        self.offset = "OFFSET <int32>$offset"
        self.add_variable("offset", offset, replace=replace)
        return self

    def set_limit(self, limit: int | None, replace: bool = False) -> "QueryBuilder":
        if limit is None:
            return self
        if self.limit is not None:
            if not replace:
                raise QueryBuilderError(
                    "A limit already exists. If you would like to replace it, pass in replace."
                )
        self.limit = "LIMIT <int32>$limit"
        self.add_variable("limit", limit, replace=replace)
        return self

    # needs to be add_filter and add_order_by

    def set_filter(
        self,
        filter: str,
        variables: dict[str, T.Any] | None = None,
        replace_filter: bool = False,
        replace_variables: bool = False,
    ) -> "QueryBuilder":
        if self.filter and not replace_filter:
            raise QueryBuilderError("Filter already exists.")
        self.add_variables(variables=variables, replace=replace_variables)
        self.filter = filter
        return self

    def set_order_by(
        self,
        order_by: str,
        variables: dict[str, T.Any] | None = None,
        replace_order_by: bool = False,
        replace_variables: bool = False,
    ) -> "QueryBuilder":
        if self.order_by and not replace_order_by:
            raise QueryBuilderError("Order by already exists.")
        self.add_variables(variables=variables, replace=replace_variables)
        self.order_by = order_by
        return self

    def set_full_query_str(
        self,
        full_query_str: str,
        replace: bool = False,
        variables: dict[str, T.Any] | None = None,
        replace_variables: bool = False,
    ) -> "QueryBuilder":
        if self.full_query_str and not replace:
            raise QueryBuilderError("full_query_str already exists.")
        self.add_variables(variables=variables, replace=replace_variables)
        pattern_to_replace = "".join(
            random.choice(string.ascii_letters + string.digits) for _ in range(10)
        )
        self.full_query_str = full_query_str.replace("$$", pattern_to_replace)
        self.pattern_to_replace = pattern_to_replace
        return self



import typing as T
from pydantic import BaseModel, ValidationInfo, model_validator
from fastgql.gql_ast.models import Node


class GQLConfigDict(T.TypedDict, total=False):
    """A TypedDict for configuring FastGQL behaviour."""

    type_name: str
    input_type_name: str
    description: str


class GQL(BaseModel):
    gql_config: T.ClassVar[GQLConfigDict] = {}

    @classmethod
    def gql_type_name(cls) -> str:
        return cls.gql_config.get("type_name", cls.__name__)

    @classmethod
    def gql_description(cls) -> str | None:
        return cls.gql_config.get("description")


class GQLInterface(GQL):
    pass


class GQLInput(GQL):
    @classmethod
    def gql_input_type_name(cls) -> str:
        return cls.gql_config.get("input_type_name", cls.__name__)

    @model_validator(mode="before")
    @classmethod
    def _to_snake_case(cls, data: T.Any, info: ValidationInfo) -> T.Any:
        if context := info.context:
            if display_to_python_map := context.get("_display_to_python_map"):
                return {display_to_python_map[k]: v for k, v in data.items()}
        return data


class GQLError(Exception):
    def __init__(
        self,
        message: str,
        *,
        node: Node | list[Node] | None = None,
        path: tuple[str, ...] | None = None,
        original_error: Exception | None = None,
        extensions: dict[str, T.Any] | None = None,
        capture_exception: bool = True,
    ):
        super().__init__(message)
        self.message = message
        self.node = node
        self.path = path
        self.original_error = original_error
        self.extensions = extensions
        self.capture_exception = capture_exception


__all__ = ["GQL", "GQLInput", "GQLConfigDict", "GQLError", "GQLInterface"]



import typing as T
from fastapi import Request, Response, BackgroundTasks
from fastgql import GQLError
from fastgql.gql_ast.models import Node


class BaseContext:
    def __init__(
        self,
        request: Request,
        response: Response,
        background_tasks: BackgroundTasks,
        errors: list[GQLError],
        variables: dict[str, T.Any],
    ):
        self.request = request
        self.response = response
        self.background_tasks = background_tasks
        self.errors = errors
        self.variables = variables

        self.overwrite_return_value_map: dict[Node, T.Any] = {}



import typing as T
from uuid import UUID
import json
from datetime import datetime, date, time
from graphql import GraphQLScalarType, ValueNode
from graphql.utilities import value_from_ast_untyped
from pydantic_core import to_jsonable_python


def serialize_datetime(value: datetime) -> str:
    return value.isoformat()


def parse_datetime_value(value: T.Any) -> datetime:
    return datetime.fromisoformat(value)


def parse_datetime_literal(
    value_node: ValueNode, variables: T.Optional[T.Dict[str, T.Any]] = None
) -> datetime:
    ast_value = value_from_ast_untyped(value_node, variables)
    return parse_datetime_value(ast_value)


def serialize_date(value: date) -> str:
    return value.isoformat()


def parse_date_value(value: T.Any) -> date:
    return date.fromisoformat(value)


def parse_date_literal(
    value_node: ValueNode, variables: T.Optional[T.Dict[str, T.Any]] = None
) -> date:
    ast_value = value_from_ast_untyped(value_node, variables)
    return parse_date_value(ast_value)


def serialize_time(value: time) -> str:
    return value.isoformat()


def parse_time_value(value: T.Any) -> time:
    return time.fromisoformat(value)


def parse_time_literal(
    value_node: ValueNode, variables: T.Optional[T.Dict[str, T.Any]] = None
) -> time:
    ast_value = value_from_ast_untyped(value_node, variables)
    return parse_time_value(ast_value)


def serialize_uuid(value: UUID) -> str:
    return str(value)


def parse_uuid_value(value: T.Any) -> UUID:
    return UUID(value)


def parse_uuid_literal(
    value_node: ValueNode, variables: T.Optional[T.Dict[str, T.Any]] = None
) -> UUID:
    ast_value = value_from_ast_untyped(value_node, variables)
    return parse_uuid_value(ast_value)


def serialize_json(value: dict) -> str:
    return json.dumps(to_jsonable_python(value))


def parse_json_value(value: T.Any) -> dict:
    return json.loads(value)


def parse_json_literal(
    value_node: ValueNode, variables: T.Optional[T.Dict[str, T.Any]] = None
) -> dict:
    ast_value = value_from_ast_untyped(value_node, variables)
    return parse_json_value(ast_value)


DateTimeScalar = GraphQLScalarType(
    name="DateTime",
    description="Datetime given as ISO.",
    serialize=serialize_datetime,
    parse_value=parse_datetime_value,
    parse_literal=parse_datetime_literal,
)

UUIDScalar = GraphQLScalarType(
    name="UUID",
    description="UUID given as String.",
    serialize=serialize_uuid,
    parse_value=parse_uuid_value,
    parse_literal=parse_uuid_literal,
)
DateScalar = GraphQLScalarType(
    name="Date",
    description="Date given as ISO.",
    serialize=serialize_date,
    parse_value=parse_date_value,
    parse_literal=parse_date_literal,
)
TimeScalar = GraphQLScalarType(
    name="Time",
    description="Time given as ISO.",
    serialize=serialize_time,
    parse_value=parse_time_value,
    parse_literal=parse_time_literal,
)
JSONScalar = GraphQLScalarType(
    name="JSON",
    description="JSON.",
    serialize=serialize_json,
    parse_value=parse_json_value,
    parse_literal=parse_json_literal,
)

__all__ = ["UUIDScalar", "DateScalar", "TimeScalar", "DateTimeScalar", "JSONScalar"]



import typing as T
from fastapi import Request, Response
from collections import OrderedDict
from dataclasses import dataclass

from pydantic import create_model
import graphql

from fastgql.info import Info, ContextType
from fastgql.gql_models import GQL, GQLError

InfoType = T.TypeVar("InfoType", bound=Info)


@dataclass
class GraphQLRequestData:
    # query is optional here as it can be added by an extensions
    # (for example an extension for persisted queries)
    query: str | None
    variables: dict[str, T.Any] | None
    operation_name: str | None


@dataclass
class Result:
    data: T.Any | None
    errors: list[graphql.GraphQLError] | None
    extensions: list[T.Any] | None


RESULT_WRAPPERS = T.Optional[
    T.List[
        T.Callable[
            [GraphQLRequestData, Result, dict[str, T.Any], Request, Response],
            T.Coroutine[T.Any, T.Any, Result],
        ]
    ]
]


def gql_errors_to_graphql_errors(
    gql_errors: list[GQLError],
) -> list[graphql.GraphQLError]:
    graphql_errors: list[graphql.GraphQLError] = []
    for e in gql_errors:
        graphql_errors.append(
            graphql.GraphQLError(
                message=e.message,
                nodes=e.node.original_node if e.node else None,
                path=e.path,
                original_error=e.original_error,
                extensions=e.extensions,
            )
        )
    return graphql_errors


class CacheDict(OrderedDict):
    """Dict with a limited length, ejecting LRUs as needed."""

    def __init__(self, *args, cache_len: int, **kwargs):
        assert cache_len > 0
        self.cache_len = cache_len

        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        super().move_to_end(key)

        while len(self) > self.cache_len:
            oldkey = next(iter(self))
            super().__delitem__(oldkey)

    def __getitem__(self, key):
        val = super().__getitem__(key)
        super().move_to_end(key)

        return val


def combine_models(name: str, *models: T.Type[GQL]) -> T.Type[GQL]:
    combined_model = create_model(name, __base__=GQL)

    seen_fields = set()
    excluded_methods = set(dir(GQL))

    for model in models:
        for field_name, field_value in model.model_fields.items():
            if field_name in seen_fields:
                raise ValueError(f"Conflicting field: {field_name}")
            setattr(combined_model, field_name, field_value)
            seen_fields.add(field_name)

        for method_name in dir(model):
            if (
                not method_name.startswith("_")
                and callable(getattr(model, method_name))
                and method_name not in excluded_methods
            ):
                method = getattr(model, method_name)
                if isinstance(model.__dict__.get(method_name), staticmethod):
                    method = staticmethod(method)
                elif isinstance(model.__dict__.get(method_name), classmethod):
                    method = classmethod(method)
                if method_name in dir(combined_model):
                    raise ValueError(f"Conflicting method: {method_name}")
                setattr(combined_model, method_name, method)

    combined_model.model_rebuild()
    return combined_model


def parse_value(variables: dict[str, T.Any] | None, v: T.Any) -> T.Any:
    if isinstance(v, dict):
        return {
            k: parse_value(variables=variables, v=inner_v) for k, inner_v in v.items()
        }
    if isinstance(v, list):
        return [parse_value(variables=variables, v=inner_v) for inner_v in v]
    if isinstance(v, graphql.VariableNode):
        if v.name.value not in variables:
            raise Exception(f"Variable {v.name.value} was not given.")
        return variables[v.name.value]
    return v


def build_is_not_nullable_map(
    schema: graphql.GraphQLSchema,
) -> dict[str, dict[str, bool]]:
    is_not_nullable_map: dict[str, dict[str, bool]] = {}
    for type_name, gql_type in schema.type_map.items():
        if isinstance(gql_type, graphql.GraphQLObjectType):
            is_not_nullable_map[type_name] = {}
            type_map = is_not_nullable_map[type_name]
            for field_name, field_val in gql_type.fields.items():
                field_val: graphql.GraphQLField
                type_map[field_name] = isinstance(
                    field_val.type, graphql.GraphQLNonNull
                )
    return is_not_nullable_map


def get_root_type(
    gql_field: graphql.GraphQLField | graphql.GraphQLObjectType,
) -> graphql.GraphQLObjectType | list[graphql.GraphQLObjectType]:
    if isinstance(gql_field, graphql.GraphQLObjectType):
        return gql_field
    t = gql_field.type
    while True:
        if isinstance(t, graphql.GraphQLUnionType):
            return [get_root_type(gql_field=tt) for tt in t.types]
        if isinstance(t, graphql.GraphQLObjectType):
            return t
        t = t.of_type


__all__ = [
    "InfoType",
    "ContextType",
    "get_root_type",
    "build_is_not_nullable_map",
    "parse_value",
    "combine_models",
    "Result",
    "gql_errors_to_graphql_errors",
    "CacheDict",
    "Info",
    "GraphQLRequestData",
    "RESULT_WRAPPERS",
]
